@@ -1,5 +1,7 @@
 #![no_std]
 #![cfg_attr(docsrs, feature(doc_cfg))]
+#![allow(clippy::new_without_default)]
+#![deny(unsafe_op_in_unsafe_fn)]
 
 //! Interoperability library for Rust Windowing applications.
 //!
@@ -10,7 +12,7 @@
 //!
 //! ## Safety guarantees
 //!
-//! Please see the docs of [`HasRawWindowHandle`] and [`HasRawDisplayHandle`].
+//! Please see the docs of [`HasWindowHandle`] and [`HasDisplayHandle`].
 //!
 //! ## Platform handle initialization
 //!
@@ -33,10 +35,9 @@ extern crate std;
 
 mod android;
 mod appkit;
-#[cfg(any(feature = "std", not(target_os = "android")))]
-#[cfg_attr(docsrs, doc(cfg(any(feature = "std", not(target_os = "android")))))]
 mod borrowed;
 mod haiku;
+mod ohos;
 mod redox;
 mod uikit;
 mod unix;
@@ -45,26 +46,27 @@ mod windows;
 
 pub use android::{AndroidDisplayHandle, AndroidNdkWindowHandle};
 pub use appkit::{AppKitDisplayHandle, AppKitWindowHandle};
-#[cfg(any(feature = "std", not(target_os = "android")))]
-pub use borrowed::{
-    Active, ActiveHandle, DisplayHandle, HandleError, HasDisplayHandle, HasWindowHandle,
-    WindowHandle,
-};
+pub use borrowed::{DisplayHandle, HasDisplayHandle, HasWindowHandle, WindowHandle};
 pub use haiku::{HaikuDisplayHandle, HaikuWindowHandle};
+pub use ohos::{OhosDisplayHandle, OhosNdkWindowHandle};
 pub use redox::{OrbitalDisplayHandle, OrbitalWindowHandle};
 pub use uikit::{UiKitDisplayHandle, UiKitWindowHandle};
 pub use unix::{
     DrmDisplayHandle, DrmWindowHandle, GbmDisplayHandle, GbmWindowHandle, WaylandDisplayHandle,
     WaylandWindowHandle, XcbDisplayHandle, XcbWindowHandle, XlibDisplayHandle, XlibWindowHandle,
 };
-pub use web::{WebDisplayHandle, WebWindowHandle};
+pub use web::{
+    WebCanvasWindowHandle, WebDisplayHandle, WebOffscreenCanvasWindowHandle, WebWindowHandle,
+};
 pub use windows::{Win32WindowHandle, WinRtWindowHandle, WindowsDisplayHandle};
 
+use core::fmt;
+
 /// Window that wraps around a raw window handle.
 ///
 /// # Safety
 ///
-/// Users can safely assume that non-`null`/`0` fields are valid handles, and it is up to the
+/// Users can safely assume that pointers and non-zero fields are valid, and it is up to the
 /// implementer of this trait to ensure that condition is upheld.
 ///
 /// Despite that qualification, implementers should still make a best-effort attempt to fill in all
@@ -74,27 +76,15 @@ pub use windows::{Win32WindowHandle, WinRtWindowHandle, WindowsDisplayHandle};
 ///
 /// The exact handles returned by `raw_window_handle` must remain consistent between multiple calls
 /// to `raw_window_handle` as long as not indicated otherwise by platform specific events.
+#[deprecated = "Use `HasWindowHandle` instead"]
 pub unsafe trait HasRawWindowHandle {
-    fn raw_window_handle(&self) -> RawWindowHandle;
+    fn raw_window_handle(&self) -> Result<RawWindowHandle, HandleError>;
 }
 
-unsafe impl<'a, T: HasRawWindowHandle + ?Sized> HasRawWindowHandle for &'a T {
-    fn raw_window_handle(&self) -> RawWindowHandle {
-        (*self).raw_window_handle()
-    }
-}
-#[cfg(feature = "alloc")]
-#[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
-unsafe impl<T: HasRawWindowHandle + ?Sized> HasRawWindowHandle for alloc::rc::Rc<T> {
-    fn raw_window_handle(&self) -> RawWindowHandle {
-        (**self).raw_window_handle()
-    }
-}
-#[cfg(feature = "alloc")]
-#[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
-unsafe impl<T: HasRawWindowHandle + ?Sized> HasRawWindowHandle for alloc::sync::Arc<T> {
-    fn raw_window_handle(&self) -> RawWindowHandle {
-        (**self).raw_window_handle()
+#[allow(deprecated)]
+unsafe impl<T: HasWindowHandle + ?Sized> HasRawWindowHandle for T {
+    fn raw_window_handle(&self) -> Result<RawWindowHandle, HandleError> {
+        self.window_handle().map(Into::into)
     }
 }
 
@@ -114,7 +104,7 @@ unsafe impl<T: HasRawWindowHandle + ?Sized> HasRawWindowHandle for alloc::sync::
 /// some hints on where this variant might be expected.
 ///
 /// Note that these "Availability Hints" are not normative. That is to say, a
-/// [`HasRawWindowHandle`] implementor is completely allowed to return something
+/// [`HasWindowHandle`] implementor is completely allowed to return something
 /// unexpected. (For example, it's legal for someone to return a
 /// [`RawWindowHandle::Xlib`] on macOS, it would just be weird, and probably
 /// requires something like XQuartz be used).
@@ -142,6 +132,11 @@ pub enum RawWindowHandle {
     /// This variant is used by the Orbital Windowing System in the Redox
     /// operating system.
     Orbital(OrbitalWindowHandle),
+    /// A raw window handle for the OpenHarmony OS NDK
+    ///
+    /// ## Availability Hints
+    /// This variant is used on OpenHarmony OS (`target_env = "ohos"`).
+    OhosNdk(OhosNdkWindowHandle),
     /// A raw window handle for Xlib.
     ///
     /// ## Availability Hints
@@ -171,7 +166,7 @@ pub enum RawWindowHandle {
     ///
     /// ## Availability Hints
     /// This variant is present regardless of windowing backend and likely to be used with
-    /// EGL_MESA_platfrom_gbm or EGL_KHR_platfrom_gbm.
+    /// EGL_MESA_platform_gbm or EGL_KHR_platform_gbm.
     Gbm(GbmWindowHandle),
     /// A raw window handle for Win32.
     ///
@@ -188,6 +183,20 @@ pub enum RawWindowHandle {
     /// ## Availability Hints
     /// This variant is used on Wasm or asm.js targets when targeting the Web/HTML5.
     Web(WebWindowHandle),
+    /// A raw window handle for a Web canvas registered via [`wasm-bindgen`].
+    ///
+    /// ## Availability Hints
+    /// This variant is used on Wasm or asm.js targets when targeting the Web/HTML5.
+    ///
+    /// [`wasm-bindgen`]: https://crates.io/crates/wasm-bindgen
+    WebCanvas(WebCanvasWindowHandle),
+    /// A raw window handle for a Web offscreen canvas registered via [`wasm-bindgen`].
+    ///
+    /// ## Availability Hints
+    /// This variant is used on Wasm or asm.js targets when targeting the Web/HTML5.
+    ///
+    /// [`wasm-bindgen`]: https://crates.io/crates/wasm-bindgen
+    WebOffscreenCanvas(WebOffscreenCanvasWindowHandle),
     /// A raw window handle for Android NDK.
     ///
     /// ## Availability Hints
@@ -204,7 +213,7 @@ pub enum RawWindowHandle {
 ///
 /// # Safety
 ///
-/// Users can safely assume that non-`null`/`0` fields are valid handles, and it is up to the
+/// Users can safely assume that pointers and non-zero fields are valid, and it is up to the
 /// implementer of this trait to ensure that condition is upheld.
 ///
 /// Despite that qualification, implementers should still make a best-effort attempt to fill in all
@@ -214,29 +223,15 @@ pub enum RawWindowHandle {
 ///
 /// The exact handles returned by `raw_display_handle` must remain consistent between multiple calls
 /// to `raw_display_handle` as long as not indicated otherwise by platform specific events.
+#[deprecated = "Use `HasDisplayHandle` instead"]
 pub unsafe trait HasRawDisplayHandle {
-    fn raw_display_handle(&self) -> RawDisplayHandle;
+    fn raw_display_handle(&self) -> Result<RawDisplayHandle, HandleError>;
 }
 
-unsafe impl<'a, T: HasRawDisplayHandle + ?Sized> HasRawDisplayHandle for &'a T {
-    fn raw_display_handle(&self) -> RawDisplayHandle {
-        (*self).raw_display_handle()
-    }
-}
-
-#[cfg(feature = "alloc")]
-#[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
-unsafe impl<T: HasRawDisplayHandle + ?Sized> HasRawDisplayHandle for alloc::rc::Rc<T> {
-    fn raw_display_handle(&self) -> RawDisplayHandle {
-        (**self).raw_display_handle()
-    }
-}
-
-#[cfg(feature = "alloc")]
-#[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
-unsafe impl<T: HasRawDisplayHandle + ?Sized> HasRawDisplayHandle for alloc::sync::Arc<T> {
-    fn raw_display_handle(&self) -> RawDisplayHandle {
-        (**self).raw_display_handle()
+#[allow(deprecated)]
+unsafe impl<T: HasDisplayHandle + ?Sized> HasRawDisplayHandle for T {
+    fn raw_display_handle(&self) -> Result<RawDisplayHandle, HandleError> {
+        self.display_handle().map(Into::into)
     }
 }
 
@@ -262,7 +257,7 @@ unsafe impl<T: HasRawDisplayHandle + ?Sized> HasRawDisplayHandle for alloc::sync
 /// some hints on where this variant might be expected.
 ///
 /// Note that these "Availability Hints" are not normative. That is to say, a
-/// [`HasRawDisplayHandle`] implementor is completely allowed to return something
+/// [`HasDisplayHandle`] implementor is completely allowed to return something
 /// unexpected. (For example, it's legal for someone to return a
 /// [`RawDisplayHandle::Xlib`] on macOS, it would just be weird, and probably
 /// requires something like XQuartz be used).
@@ -290,6 +285,11 @@ pub enum RawDisplayHandle {
     /// This variant is used by the Orbital Windowing System in the Redox
     /// operating system.
     Orbital(OrbitalDisplayHandle),
+    /// A raw display handle for OpenHarmony OS NDK
+    ///
+    /// ## Availability Hints
+    /// This variant is used on OpenHarmony OS (`target_env = "ohos"`).
+    Ohos(OhosDisplayHandle),
     /// A raw display handle for Xlib.
     ///
     /// ## Availability Hints
@@ -319,7 +319,7 @@ pub enum RawDisplayHandle {
     ///
     /// ## Availability Hints
     /// This variant is present regardless of windowing backend and likely to be used with
-    /// EGL_MESA_platfrom_gbm or EGL_KHR_platfrom_gbm.
+    /// EGL_MESA_platform_gbm or EGL_KHR_platform_gbm.
     Gbm(GbmDisplayHandle),
     /// A raw display handle for Win32.
     ///
@@ -343,6 +343,53 @@ pub enum RawDisplayHandle {
     Haiku(HaikuDisplayHandle),
 }
 
+/// An error that can occur while fetching a display or window handle.
+#[derive(Debug, Clone)]
+#[non_exhaustive]
+pub enum HandleError {
+    /// The underlying handle cannot be represented using the types in this crate.
+    ///
+    /// This may be returned if the underlying window system does not support any of the
+    /// representative C window handles in this crate. For instance, if you were using a pure Rust
+    /// library to set up X11 (like [`x11rb`]), you would not be able to use any of the
+    /// [`RawWindowHandle`] variants, as they all represent C types.
+    ///
+    /// Another example would be a system that isn't supported by `raw-window-handle` yet,
+    /// like some game consoles.
+    ///
+    /// In the event that this error is returned, you should try to use the underlying window
+    /// system's native API to get the handle you need.
+    ///
+    /// [`x11rb`]: https://crates.io/crates/x11rb
+    NotSupported,
+
+    /// The underlying handle is not available.
+    ///
+    /// In some cases the underlying window handle may become temporarily unusable. For example, on
+    /// Android, the native window pointer can arbitrarily be replaced or removed by the system. In
+    /// this case, returning a window handle would be disingenuous, as it would not be usable. A
+    /// similar situation can occur on Wayland for the layer shell windows.
+    ///
+    /// In the event that this error is returned, you should wait until the handle becomes available
+    /// again.
+    Unavailable,
+}
+
+impl fmt::Display for HandleError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Self::NotSupported => write!(
+                f,
+                "the underlying handle cannot be represented using the types in this crate"
+            ),
+            Self::Unavailable => write!(f, "the underlying handle is not available"),
+        }
+    }
+}
+
+#[cfg(feature = "std")]
+impl std::error::Error for HandleError {}
+
 macro_rules! from_impl {
     ($($to:ident, $enum:ident, $from:ty)*) => ($(
         impl From<$from> for $to {
@@ -356,6 +403,7 @@ macro_rules! from_impl {
 from_impl!(RawDisplayHandle, UiKit, UiKitDisplayHandle);
 from_impl!(RawDisplayHandle, AppKit, AppKitDisplayHandle);
 from_impl!(RawDisplayHandle, Orbital, OrbitalDisplayHandle);
+from_impl!(RawDisplayHandle, Ohos, OhosDisplayHandle);
 from_impl!(RawDisplayHandle, Xlib, XlibDisplayHandle);
 from_impl!(RawDisplayHandle, Xcb, XcbDisplayHandle);
 from_impl!(RawDisplayHandle, Wayland, WaylandDisplayHandle);
@@ -369,6 +417,7 @@ from_impl!(RawDisplayHandle, Haiku, HaikuDisplayHandle);
 from_impl!(RawWindowHandle, UiKit, UiKitWindowHandle);
 from_impl!(RawWindowHandle, AppKit, AppKitWindowHandle);
 from_impl!(RawWindowHandle, Orbital, OrbitalWindowHandle);
+from_impl!(RawWindowHandle, OhosNdk, OhosNdkWindowHandle);
 from_impl!(RawWindowHandle, Xlib, XlibWindowHandle);
 from_impl!(RawWindowHandle, Xcb, XcbWindowHandle);
 from_impl!(RawWindowHandle, Wayland, WaylandWindowHandle);
@@ -377,5 +426,74 @@ from_impl!(RawWindowHandle, Gbm, GbmWindowHandle);
 from_impl!(RawWindowHandle, Win32, Win32WindowHandle);
 from_impl!(RawWindowHandle, WinRt, WinRtWindowHandle);
 from_impl!(RawWindowHandle, Web, WebWindowHandle);
+from_impl!(RawWindowHandle, WebCanvas, WebCanvasWindowHandle);
+from_impl!(
+    RawWindowHandle,
+    WebOffscreenCanvas,
+    WebOffscreenCanvasWindowHandle
+);
 from_impl!(RawWindowHandle, AndroidNdk, AndroidNdkWindowHandle);
 from_impl!(RawWindowHandle, Haiku, HaikuWindowHandle);
+
+#[cfg(test)]
+mod tests {
+    use core::panic::{RefUnwindSafe, UnwindSafe};
+    use static_assertions::{assert_impl_all, assert_not_impl_any};
+
+    use super::*;
+
+    #[test]
+    fn auto_traits() {
+        assert_impl_all!(RawDisplayHandle: UnwindSafe, RefUnwindSafe, Unpin);
+        assert_not_impl_any!(RawDisplayHandle: Send, Sync);
+        assert_impl_all!(DisplayHandle<'_>: UnwindSafe, RefUnwindSafe, Unpin);
+        assert_not_impl_any!(DisplayHandle<'_>: Send, Sync);
+        assert_impl_all!(RawWindowHandle: UnwindSafe, RefUnwindSafe, Unpin);
+        assert_not_impl_any!(RawWindowHandle: Send, Sync);
+        assert_impl_all!(WindowHandle<'_>: UnwindSafe, RefUnwindSafe, Unpin);
+        assert_not_impl_any!(WindowHandle<'_>: Send, Sync);
+        assert_impl_all!(HandleError: Send, Sync, UnwindSafe, RefUnwindSafe, Unpin);
+
+        // TODO: Unsure if some of these should not actually be Send + Sync
+        assert_impl_all!(UiKitDisplayHandle: Send, Sync);
+        assert_impl_all!(AppKitDisplayHandle: Send, Sync);
+        assert_impl_all!(OrbitalDisplayHandle: Send, Sync);
+        assert_impl_all!(OhosDisplayHandle: Send, Sync);
+        assert_not_impl_any!(XlibDisplayHandle: Send, Sync);
+        assert_not_impl_any!(XcbDisplayHandle: Send, Sync);
+        assert_not_impl_any!(WaylandDisplayHandle: Send, Sync);
+        assert_impl_all!(DrmDisplayHandle: Send, Sync);
+        assert_not_impl_any!(GbmDisplayHandle: Send, Sync);
+        assert_impl_all!(WindowsDisplayHandle: Send, Sync);
+        assert_impl_all!(WebDisplayHandle: Send, Sync);
+        assert_impl_all!(AndroidDisplayHandle: Send, Sync);
+        assert_impl_all!(HaikuDisplayHandle: Send, Sync);
+
+        // TODO: Unsure if some of these should not actually be Send + Sync
+        assert_not_impl_any!(UiKitWindowHandle: Send, Sync);
+        assert_not_impl_any!(AppKitWindowHandle: Send, Sync);
+        assert_not_impl_any!(OrbitalWindowHandle: Send, Sync);
+        assert_not_impl_any!(OhosNdkWindowHandle: Send, Sync);
+        assert_impl_all!(XlibWindowHandle: Send, Sync);
+        assert_impl_all!(XcbWindowHandle: Send, Sync);
+        assert_not_impl_any!(WaylandWindowHandle: Send, Sync);
+        assert_impl_all!(DrmWindowHandle: Send, Sync);
+        assert_not_impl_any!(GbmWindowHandle: Send, Sync);
+        assert_impl_all!(Win32WindowHandle: Send, Sync);
+        assert_not_impl_any!(WinRtWindowHandle: Send, Sync);
+        assert_impl_all!(WebWindowHandle: Send, Sync);
+        assert_not_impl_any!(WebCanvasWindowHandle: Send, Sync);
+        assert_not_impl_any!(WebOffscreenCanvasWindowHandle: Send, Sync);
+        assert_not_impl_any!(AndroidNdkWindowHandle: Send, Sync);
+        assert_not_impl_any!(HaikuWindowHandle: Send, Sync);
+    }
+
+    #[allow(deprecated, unused)]
+    fn assert_object_safe(
+        _: &dyn HasRawWindowHandle,
+        _: &dyn HasRawDisplayHandle,
+        _: &dyn HasWindowHandle,
+        _: &dyn HasDisplayHandle,
+    ) {
+    }
+}
